@@ -27,6 +27,8 @@
 #include "xua_msg.h"
 #include <osmocom/sigtran/protocol/sua.h>
 
+#include "ss7_as.h"
+#include "ss7_asp.h"
 #include "ss7_internal.h"
 #include "ss7_xua_srv.h"
 #include "xua_asp_fsm.h"
@@ -42,9 +44,6 @@
  * *
  */
 
-/* According to RFC3868 Section 8 */
-#define XUA_T_ACK_SEC	2
-
 static const struct value_string xua_asp_event_names[] = {
 	{ XUA_ASP_E_M_ASP_UP_REQ,	"M-ASP_UP.req" },
 	{ XUA_ASP_E_M_ASP_ACTIVE_REQ,	"M-ASP_ACTIVE.req" },
@@ -86,7 +85,7 @@ struct xua_asp_fsm_priv {
 	/* routing context[s]: list of 32bit integers */
 	/* ACTIVE: traffic mode type, tid label, drn label ? */
 
-	struct {
+	struct { /* RFC3868 & RFC4666 timer T(ack) */
 		struct osmo_timer_list timer;
 		int out_event;
 	} t_ack;
@@ -171,32 +170,14 @@ static int determine_traf_mode(struct osmo_ss7_asp *asp)
 /* add M3UA_IEI_ROUTE_CTX to xua_msg containig all routing keys of ASs within ASP */
 static int xua_msg_add_asp_rctx(struct xua_msg *xua, struct osmo_ss7_asp *asp)
 {
-	struct osmo_ss7_as *as;
-	uint32_t rctx[128];
-	unsigned int i = 0;
-
-	/* iterate over all ASs and build array of routing contexts */
-	llist_for_each_entry(as, &asp->inst->as_list, list) {
-		if (!osmo_ss7_as_has_asp(as, asp))
-			continue;
-		rctx[i++] = htonl(as->cfg.routing_key.context);
-		if (i >= ARRAY_SIZE(rctx)-1) {
-			break;
-		}
-	}
-	/* add xUA IE with routing contests to the message (if any) */
-	if (i) {
-		/* bail out (and not add the IE) if there's only one routing context (and hence
-		 * only one AS) within this ASP, and that routing context is zero, meaning no routing
-		 * context IE shall be used */
-		if (i == 1 && rctx[0] == 0)
-			return 0;
-
-		xua_msg_add_data(xua, M3UA_IEI_ROUTE_CTX, i*sizeof(uint32_t), (uint8_t *)rctx);
-	}
+	uint32_t rctx[OSMO_SS7_MAX_RCTX_COUNT];
+	unsigned int cnt;
 
+	cnt = ss7_asp_get_all_rctx_be(asp, rctx, ARRAY_SIZE(rctx), NULL);
+	if (cnt > 0)
+		xua_msg_add_data(xua, M3UA_IEI_ROUTE_CTX, cnt*sizeof(uint32_t), (uint8_t *)rctx);
 	/* return count of routing contexts added */
-	return i;
+	return cnt;
 }
 
 /* ask the xUA implementation to transmit a specific message */
@@ -213,13 +194,33 @@ static int peer_send(struct osmo_fsm_inst *fi, int out_event, struct xua_msg *in
 		/* RFC 3868 Ch. 3.5.1 */
 		xua->hdr = XUA_HDR(SUA_MSGC_ASPSM, SUA_ASPSM_UP);
 		/* Optional: ASP ID */
-		if (asp->asp_id_present)
-			xua_msg_add_u32(xua, SUA_IEI_ASP_ID, asp->asp_id);
+#if 0
+		/* TODO: RFC 4666 3.8.1:
+		 * 'The "ASP Identifier Required" error is sent by an SGP in
+		 * response to an ASP Up message that does not contain an ASP
+		 * Identifier parameter when the SGP requires one. The ASP SHOULD
+		 * resend the ASP Up message with an ASP Identifier.' */
+		if (ss7_asp_peer_requires_asp_id(asp)) { /* Maybe configure in VTY "asp" node? */
+			asp_id = /* get a unique id of asp within as, eg. the index in as->asps[] */;
+			xua_msg_add_u32(xua, SUA_IEI_ASP_ID, asp_id);
+		}
+#endif
 		/* Optional: Info String */
 		break;
 	case XUA_ASP_E_ASPSM_ASPUP_ACK:
 		/* RFC3868 Ch. 3.5.2 */
 		xua->hdr = XUA_HDR(SUA_MSGC_ASPSM, SUA_ASPSM_UP_ACK);
+		/* Optional: ASP ID */
+#if 0
+		/* TODO: RFC 4666 3.5.2:
+		 * "The optional ASP Identifier parameter is specifically useful for IPSP
+		 * communication.  In that case, the IPSP answering the ASP Up message
+		 * MAY include its own ASP Identifier value." */
+		if (ss7_asp_peer_requires_asp_id(asp)) { /* Maybe configure in VTY "asp" node? */
+			asp_id = /* get a unique id of asp within as, eg. the index in as->asps[] */;
+			xua_msg_add_u32(xua, SUA_IEI_ASP_ID, asp_id);
+		}
+#endif
 		/* Optional: Info String */
 		break;
 	case XUA_ASP_E_ASPSM_ASPDN:
@@ -313,6 +314,7 @@ static void xua_t_ack_cb(void *data)
 {
 	struct osmo_fsm_inst *fi = data;
 	struct xua_asp_fsm_priv *xafp = fi->priv;
+	uint32_t timeout_sec;
 
 	LOGPFSML(fi, LOGL_INFO, "T(ack) callback: re-transmitting event %s\n",
 		osmo_fsm_event_name(fi->fsm, xafp->t_ack.out_event));
@@ -321,13 +323,15 @@ static void xua_t_ack_cb(void *data)
 	peer_send(fi, xafp->t_ack.out_event, NULL);
 
 	/* Re-start the timer */
-	osmo_timer_schedule(&xafp->t_ack.timer, XUA_T_ACK_SEC, 0);
+	timeout_sec = osmo_tdef_get(xafp->asp->cfg.T_defs_xua, SS7_ASP_XUA_T_ACK, OSMO_TDEF_S, -1);
+	osmo_timer_schedule(&xafp->t_ack.timer, timeout_sec, 0);
 }
 
 static int peer_send_and_start_t_ack(struct osmo_fsm_inst *fi,
 				     int out_event)
 {
 	struct xua_asp_fsm_priv *xafp = fi->priv;
+	uint32_t timeout_sec;
 	int rc;
 
 	rc = peer_send(fi, out_event, NULL);
@@ -338,7 +342,8 @@ static int peer_send_and_start_t_ack(struct osmo_fsm_inst *fi,
 	xafp->t_ack.timer.cb = xua_t_ack_cb,
 	xafp->t_ack.timer.data = fi;
 
-	osmo_timer_schedule(&xafp->t_ack.timer, XUA_T_ACK_SEC, 0);
+	timeout_sec = osmo_tdef_get(xafp->asp->cfg.T_defs_xua, SS7_ASP_XUA_T_ACK, OSMO_TDEF_S, -1);
+	osmo_timer_schedule(&xafp->t_ack.timer, timeout_sec, 0);
 
 	return rc;
 }
@@ -353,7 +358,7 @@ static const uint32_t evt_ack_map[_NUM_XUA_ASP_E] = {
 
 /* Helper function to dispatch an ASP->AS event to all AS of which this
  * ASP is a memmber.  Ignores routing contexts for now. */
-static void dispatch_to_all_as(struct osmo_fsm_inst *fi, uint32_t event)
+static void dispatch_to_all_as(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
 	struct xua_asp_fsm_priv *xafp = fi->priv;
 	struct osmo_ss7_asp *asp = xafp->asp;
@@ -363,7 +368,7 @@ static void dispatch_to_all_as(struct osmo_fsm_inst *fi, uint32_t event)
 	llist_for_each_entry(as, &inst->as_list, list) {
 		if (!osmo_ss7_as_has_asp(as, asp))
 			continue;
-		osmo_fsm_inst_dispatch(as->fi, event, asp);
+		osmo_fsm_inst_dispatch(as->fi, event, data);
 	}
 }
 
@@ -416,7 +421,9 @@ static void check_stop_t_ack(struct osmo_fsm_inst *fi, uint32_t event)
 
 static void xua_asp_fsm_down_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
 {
-	dispatch_to_all_as(fi, XUA_ASPAS_ASP_DOWN_IND);
+	struct xua_asp_fsm_priv *xafp = fi->priv;
+	struct osmo_ss7_asp *asp = xafp->asp;
+	dispatch_to_all_as(fi, XUA_ASPAS_ASP_DOWN_IND, asp);
 }
 
 static void xua_asp_fsm_down(struct osmo_fsm_inst *fi, uint32_t event, void *data)
@@ -437,6 +444,11 @@ static void xua_asp_fsm_down(struct osmo_fsm_inst *fi, uint32_t event, void *dat
 	case XUA_ASP_E_ASPSM_ASPUP_ACK:
 		/* only if role ASP */
 		ENSURE_ASP_OR_IPSP(fi, event);
+		/* Optional ASP Identifier */
+		if ((asp_id_ie = xua_msg_find_tag(data, SUA_IEI_ASP_ID))) {
+			asp->remote_asp_id = xua_msg_part_get_u32(asp_id_ie);
+			asp->remote_asp_id_present = true;
+		}
 		osmo_fsm_inst_state_chg(fi, XUA_ASP_S_INACTIVE, 0, 0);
 		/* inform layer manager */
 		send_xlm_prim_simple(fi, OSMO_XLM_PRIM_M_ASP_UP, PRIM_OP_CONFIRM);
@@ -448,11 +460,10 @@ static void xua_asp_fsm_down(struct osmo_fsm_inst *fi, uint32_t event, void *dat
 	case XUA_ASP_E_ASPSM_ASPUP:
 		/* only if role SG */
 		ENSURE_SG_OR_IPSP(fi, event);
-		asp_id_ie = xua_msg_find_tag(data, SUA_IEI_ASP_ID);
 		/* Optional ASP Identifier: Store for NTFY */
-		if (asp_id_ie) {
-			asp->asp_id = xua_msg_part_get_u32(asp_id_ie);
-			asp->asp_id_present = true;
+		if ((asp_id_ie = xua_msg_find_tag(data, SUA_IEI_ASP_ID))) {
+			asp->remote_asp_id = xua_msg_part_get_u32(asp_id_ie);
+			asp->remote_asp_id_present = true;
 		}
 		/* send ACK */
 		peer_send(fi, XUA_ASP_E_ASPSM_ASPUP_ACK, NULL);
@@ -476,7 +487,20 @@ static void xua_asp_fsm_down(struct osmo_fsm_inst *fi, uint32_t event, void *dat
 
 static void xua_asp_fsm_inactive_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
 {
-	dispatch_to_all_as(fi, XUA_ASPAS_ASP_INACTIVE_IND);
+	struct xua_asp_fsm_priv *xafp = fi->priv;
+	struct osmo_ss7_asp *asp = xafp->asp;
+	/* RFC4666 4.3.4.5: "When an ASP moves from ASP-DOWN to ASP-INACTIVE within a
+	 * particular AS, a Notify message SHOULD be sent, by the ASP-UP receptor,
+	 * after sending the ASP-UP-ACK, in order to inform the ASP of the current AS
+	 * state."
+	 * NOTIFY is only transmitted by roles SG and IPSP.
+	 */
+	struct xua_as_event_asp_inactive_ind_pars pars = {
+		.asp = asp,
+		.asp_requires_notify = (asp->cfg.role != OSMO_SS7_ASP_ROLE_ASP) &&
+				       (prev_state == XUA_ASP_S_DOWN),
+	};
+	dispatch_to_all_as(fi, XUA_ASPAS_ASP_INACTIVE_IND, &pars);
 }
 
 static void xua_asp_fsm_inactive(struct osmo_fsm_inst *fi, uint32_t event, void *data)
@@ -594,7 +618,9 @@ static void xua_asp_fsm_inactive(struct osmo_fsm_inst *fi, uint32_t event, void
 
 static void xua_asp_fsm_active_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
 {
-	dispatch_to_all_as(fi, XUA_ASPAS_ASP_ACTIVE_IND);
+	struct xua_asp_fsm_priv *xafp = fi->priv;
+	struct osmo_ss7_asp *asp = xafp->asp;
+	dispatch_to_all_as(fi, XUA_ASPAS_ASP_ACTIVE_IND, asp);
 }
 
 static void xua_asp_fsm_active(struct osmo_fsm_inst *fi, uint32_t event, void *data)
@@ -712,7 +738,13 @@ static void xua_asp_fsm_cleanup(struct osmo_fsm_inst *fi, enum osmo_fsm_term_cau
 {
 	struct xua_asp_fsm_priv *xafp = fi->priv;
 
+	if (!xafp)
+		return;
+
 	osmo_timer_del(&xafp->t_ack.timer);
+
+	if (xafp->asp)
+		xafp->asp->fi = NULL;
 }
 
 static const struct osmo_fsm_state xua_asp_states[] = {
@@ -777,16 +809,16 @@ struct osmo_fsm xua_asp_fsm = {
 	.cleanup = xua_asp_fsm_cleanup,
 };
 
-static struct osmo_fsm_inst *ipa_asp_fsm_start(struct osmo_ss7_asp *asp,
-					       enum osmo_ss7_asp_role role, int log_level);
+static int ipa_asp_fsm_start(struct osmo_ss7_asp *asp,
+			     enum osmo_ss7_asp_role role, int log_level);
 
-/*! \brief Start a new ASP finite stae machine for given ASP
+/*! \brief Start a new ASP finite state machine for given ASP (stored in asp->fi)
  *  \param[in] asp Application Server Process for which to start FSM
  *  \param[in] role Role (ASP, SG, IPSP) of this FSM
  *  \param[in] log_level Logging Level for ASP FSM logging
- *  \returns FSM instance on success; NULL on error */
-struct osmo_fsm_inst *xua_asp_fsm_start(struct osmo_ss7_asp *asp,
-					enum osmo_ss7_asp_role role, int log_level)
+ *  \returns 0 on success; negative on error */
+int xua_asp_fsm_start(struct osmo_ss7_asp *asp,
+		      enum osmo_ss7_asp_role role, int log_level)
 {
 	struct osmo_fsm_inst *fi;
 	struct xua_asp_fsm_priv *xafp;
@@ -796,18 +828,23 @@ struct osmo_fsm_inst *xua_asp_fsm_start(struct osmo_ss7_asp *asp,
 
 	/* allocate as child of AS? */
 	fi = osmo_fsm_inst_alloc(&xua_asp_fsm, asp, NULL, log_level, asp->cfg.name);
+	if (!fi)
+		return -EINVAL;
 
 	xafp = talloc_zero(fi, struct xua_asp_fsm_priv);
 	if (!xafp) {
 		osmo_fsm_inst_term(fi, OSMO_FSM_TERM_ERROR, NULL);
-		return NULL;
+		return -ENOMEM;
 	}
 	xafp->role = role;
 	xafp->asp = asp;
 
 	fi->priv = xafp;
 
-	return fi;
+	/* Attach FSM to ASP: */
+	asp->fi = fi;
+
+	return 0;
 }
 
 
@@ -933,7 +970,7 @@ static void ipa_asp_fsm_wait_id_resp(struct osmo_fsm_inst *fi, uint32_t event, v
 				iafp->ipa_unit->unit_name);
 			goto out_err;
 		}
-		osmo_ss7_as_add_asp(as, asp->cfg.name);
+		ss7_as_add_asp(as, asp);
 		/* TODO: OAP Authentication? */
 		/* Send ID_ACK */
 		if (fd >= 0) {
@@ -1036,8 +1073,14 @@ static void ipa_asp_fsm_wait_id_ack(struct osmo_fsm_inst *fi, uint32_t event, vo
 
 static void ipa_asp_fsm_active_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
 {
-	dispatch_to_all_as(fi, XUA_ASPAS_ASP_INACTIVE_IND);
-	dispatch_to_all_as(fi, XUA_ASPAS_ASP_ACTIVE_IND);
+	struct ipa_asp_fsm_priv *iafp = fi->priv;
+	struct osmo_ss7_asp *asp = iafp->asp;
+	struct xua_as_event_asp_inactive_ind_pars pars = {
+		.asp = asp,
+		.asp_requires_notify = false,
+	};
+	dispatch_to_all_as(fi, XUA_ASPAS_ASP_INACTIVE_IND, &pars);
+	dispatch_to_all_as(fi, XUA_ASPAS_ASP_ACTIVE_IND, asp);
 }
 
 /* Server + Client: We're actively transmitting user data */
@@ -1053,7 +1096,12 @@ static void ipa_asp_fsm_active(struct osmo_fsm_inst *fi, uint32_t event, void *d
 
 static void ipa_asp_fsm_inactive_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
 {
-	dispatch_to_all_as(fi, XUA_ASPAS_ASP_INACTIVE_IND);
+	struct ipa_asp_fsm_priv *iafp = fi->priv;
+	struct xua_as_event_asp_inactive_ind_pars pars = {
+		.asp = iafp->asp,
+		.asp_requires_notify = false,
+	};
+	dispatch_to_all_as(fi, XUA_ASPAS_ASP_INACTIVE_IND, &pars);
 }
 
 static void ipa_asp_fsm_inactive(struct osmo_fsm_inst *fi, uint32_t event, void *data)
@@ -1065,6 +1113,36 @@ static void ipa_asp_fsm_inactive(struct osmo_fsm_inst *fi, uint32_t event, void
 	}
 }
 
+/* Assign a 4 bit SLS (as unqiue as possible) for incoming IPA PDUs.*/
+static void _ipa_asp_pick_unused_sls(struct osmo_ss7_asp *asp, const struct osmo_ss7_as *as)
+{
+	for (unsigned int sls = 0; sls <= 0x0f; sls++) {
+		bool used = false;
+		for (unsigned i = 0; i < ARRAY_SIZE(as->cfg.asps); i++) {
+			if (!as->cfg.asps[i])
+				continue;
+			if (as->cfg.asps[i] == asp)
+				continue;
+			if (!as->cfg.asps[i]->ipa.sls_assigned)
+				continue;
+			if (as->cfg.asps[i]->ipa.sls == sls) {
+				used = true;
+				break;
+			}
+		}
+		if (used)
+			continue;
+		/* Found an unused SLS, use it: */
+		asp->ipa.sls = sls;
+		asp->ipa.sls_assigned = true;
+		LOGPASP(asp, DLSS7, LOGL_DEBUG, "Assigned unsued SLS = %u\n", sls);
+		return;
+	}
+	LOGPASP(asp, DLSS7, LOGL_INFO, "All SLS in IPA AS picked, unique SLS not possible, picking random one\n");
+	asp->ipa.sls = rand() & 0x0f;
+	asp->ipa.sls_assigned = true;
+}
+
 static void ipa_asp_allstate(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
 	struct ipa_asp_fsm_priv *iafp = fi->priv;
@@ -1091,6 +1169,9 @@ static void ipa_asp_allstate(struct osmo_fsm_inst *fi, uint32_t event, void *dat
 	case XUA_ASP_E_AS_ASSIGNED:
 		as = data;
 		osmo_talloc_replace_string(iafp->ipa_unit, &iafp->ipa_unit->unit_name, as->cfg.name);
+		/* Now that AS is known, try picking an unused SLS inside the AS.
+		 * It will be applied to PDUs received from the IPA socket. */
+		_ipa_asp_pick_unused_sls(iafp->asp, as);
 		/* Now that the AS is known, start the client side: */
 		if (iafp->role == OSMO_SS7_ASP_ROLE_ASP && fi->state == IPA_ASP_S_DOWN) {
 			LOGPFSML(fi, LOGL_NOTICE, "Bringing up ASP now once it has been assigned to an AS\n");
@@ -1122,6 +1203,14 @@ static int ipa_asp_fsm_timer_cb(struct osmo_fsm_inst *fi)
 	return -1;
 }
 
+static void ipa_asp_fsm_cleanup(struct osmo_fsm_inst *fi, enum osmo_fsm_term_cause cause)
+{
+	struct ipa_asp_fsm_priv *iafp = fi->priv;
+
+	if (iafp && iafp->asp)
+		iafp->asp->fi = NULL;
+}
+
 static const struct osmo_fsm_state ipa_asp_states[] = {
 	[IPA_ASP_S_DOWN] = {
 		.in_event_mask = S(XUA_ASP_E_M_ASP_UP_REQ) |
@@ -1198,16 +1287,17 @@ struct osmo_fsm ipa_asp_fsm = {
 			       S(XUA_ASP_E_ASPSM_BEAT_ACK) |
 			       S(XUA_ASP_E_AS_ASSIGNED),
 	.allstate_action = ipa_asp_allstate,
+	.cleanup = ipa_asp_fsm_cleanup,
 };
 
 
-/*! \brief Start a new ASP finite state machine for given ASP
+/*! \brief Start a new ASP finite state machine for given ASP (stored on asp->fi)
  *  \param[in] asp Application Server Process for which to start FSM
  *  \param[in] role Role (ASP, SG, IPSP) of this FSM
  *  \param[in] log_level Logging Level for ASP FSM logging
- *  \returns FSM instance on success; NULL on error */
-static struct osmo_fsm_inst *ipa_asp_fsm_start(struct osmo_ss7_asp *asp,
-					       enum osmo_ss7_asp_role role, int log_level)
+ *  \returns 0 on success; negative on error */
+static int ipa_asp_fsm_start(struct osmo_ss7_asp *asp,
+			     enum osmo_ss7_asp_role role, int log_level)
 {
 	struct osmo_fsm_inst *fi;
 	struct ipa_asp_fsm_priv *iafp;
@@ -1217,19 +1307,24 @@ static struct osmo_fsm_inst *ipa_asp_fsm_start(struct osmo_ss7_asp *asp,
 
 	/* allocate as child of AS? */
 	fi = osmo_fsm_inst_alloc(&ipa_asp_fsm, asp, NULL, log_level, asp->cfg.name);
+	if (!fi)
+		return -EINVAL;
 
 	iafp = talloc_zero(fi, struct ipa_asp_fsm_priv);
 	if (!iafp) {
 		osmo_fsm_inst_term(fi, OSMO_FSM_TERM_ERROR, NULL);
-		return NULL;
+		return -ENOMEM;
 	}
 
 	if (as) {
 		unit_name = as->cfg.name;
+		/* Allocacate potentially unique SLS within AS since AS is already known: */
+		_ipa_asp_pick_unused_sls(asp, as);
 	} else if (asp->dyn_allocated) {
 		LOGPFSML(fi, LOGL_INFO, "Dynamic ASP is not assigned to any AS, "
 			 "using ASP name instead of AS name as ipa_unit_name\n");
 		unit_name = asp->cfg.name;
+		/* asp->ipa.sls will be assigned together with AS unit_name during XUA_ASP_E_AS_ASSIGNED. */
 	} else {
 		/* ASP in client mode will be brought up when this ASP is added
 		 * to an AS, see XUA_ASP_E_AS_ASSIGNED. */
@@ -1238,6 +1333,7 @@ static struct osmo_fsm_inst *ipa_asp_fsm_start(struct osmo_ss7_asp *asp,
 			can_start = false;
 		}
 		unit_name = asp->cfg.name;
+		/* asp->ipa.sls will be assigned together with AS unit_name during XUA_ASP_E_AS_ASSIGNED. */
 	}
 
 	iafp->role = role;
@@ -1249,8 +1345,11 @@ static struct osmo_fsm_inst *ipa_asp_fsm_start(struct osmo_ss7_asp *asp,
 
 	fi->priv = iafp;
 
+	/* Attach FSM to ASP: */
+	asp->fi = fi;
+
 	if (can_start && role == OSMO_SS7_ASP_ROLE_ASP)
 		osmo_fsm_inst_dispatch(fi, XUA_ASP_E_M_ASP_UP_REQ, NULL);
 
-	return fi;
+	return 0;
 }
